// File: FW_FLT-4.ino
// Firmware for FLT-4
// Also used on the TPA TestTool 
// Target: Arduino Nano
// For pin use and messages see the Messages document
//  
// By jan@breem.nl

#define VERSIONSTRING "FLT-4 2019-01-21"

// The following #defines are for testing the testsoftware.

// Uncomment to let the firmware think it runs on a real FLT board
// although it is on the TestTool-TPA
//#define FakeFLT

// Uncomment to make the firmware run in Local mode
// although communicating
//#define FakeLOCAL

// Uncomment to check period and duration of timer interrupt handling on COMled pin
// Should be 1 kHz and ~100 usec duration 
//#define DIAGNOSE_TIMER

// input pin definitions
#define MainsOKpin    ((PIND & 0x04) == 0x04)
#define Crydompin     ((PIND & 0x08) == 0x08)
#define Fault_Bpin    ((PINB & 0x01) == 0)
#define ClipOTW_Bpin  ((PINB & 0x02) == 0)
#define Fault_MTpin   ((PINB & 0x04) == 0)
#define ClipOTW_MTpin ((PINB & 0x08) == 0)
#define TPAisInReset  ((PINB & 0x10) == 0)

// output pin definitions
#define RSTTPAON       {PORTB &= ~0x10; PORTB |=  0x20;}
#define RSTTPAOFF      {PORTB |=  0x10; PORTB &= ~0x20;}
#define ERRORledON     PORTD &= ~0x20; 
#define ERRORledOFF    PORTD |=  0x20;
#define ERRORledTOGGLE PIND  |=  0x20;
#define COMledON       PORTD &= ~0x40;
#define COMledOFF      PORTD |=  0x40;
#define COMledTOGGLE   PIND  |=  0x40;
#define CPUledON       PORTD &= ~0x80;
#define CPUledOFF      PORTD |=  0x80;
#define CPUledTOGGLE   PIND  |=  0x80;
#define CrydomON       PORTD |=  0x08;
#define CrydomOFF      PORTD &= ~0x08;  

// for Messaging
#define INMESSAGESIZE 3
char InMessageBuffer[INMESSAGESIZE];
#define OUTMESSAGESIZE 13
char OutMessageBuffer[OUTMESSAGESIZE];

// for Volume Control
byte Command, Bas, Mid;
const int BasAddress = 0; // EEPROM addresses
const int MidAddress = 1;

// for Monitoring
byte CounterMainsOK; // Set at mains zero crossing, decremented at 1 kHz rate
int CounterCommunicationTimeout; // Set at Communication event, decremented at 1 kHz rate
byte ClipCountMT, ClipCountB; // counter increments at Clip-OTW interrupts
int AdcRef;
int Adc50, Adc12, Adc5A, AdcMin5A; // ADC readings
float V5D, V50, V12, V5A, Vmin5A;
volatile bool MainsOK, SuppliesOK, FaultMT, FaultB, OTWB, OTWMT;
int CounterOTW;
 
// General
bool Local; // When no communication is present.
bool SendSpontaneousMessage, SendExtraMessage; 
bool TestTool; // We are running on the TestTool

//************************************************************************************************
ISR(TIMER1_COMPA_vect) // called at 1 kHz rate
{
  static int CounterLedTest = 2000;
  static bool LedTest = true;
  static int CounterCpuLed, CpuLedReload, CounterErrorLed, ErrorLedReload;
  static int CounterSpontaneousMessage;
  static byte Channel;
   
#ifdef DIAGNOSE_TIMER
  COMledON // Check period and duration of timer interrupt handling  
#endif  

// Led test at startup lasts 2 seconds
  if(LedTest)
  { 
    if(CounterLedTest > 0) {CounterLedTest--; return;}
    else 
    {
      CPUledOFF
      COMledOFF
      ERRORledOFF
      LedTest = false; 
    }
  }

// Check conditions for ErrorLed
  ErrorLedReload = -1; // ErrorLedOFF
  if(!SuppliesOK) ErrorLedReload = -2; // ErrorLed continuously on
  FaultB = Fault_Bpin;
  if(FaultB) ErrorLedReload = 1000; // slow blink 
  FaultMT = Fault_MTpin;
  if(FaultMT) ErrorLedReload = 300; // fast blink
  if(FaultMT && FaultB) ErrorLedReload = 100; // very fast blink

// Test CR+ voltage from Control Amplifier 
   MainsOK = (Crydompin);   

// Check Mains Present
// CounterMainsOK is set to 15 at each mains zero crossing
  if(CounterMainsOK > 0) CounterMainsOK--;
  else MainsOK = false;
      
// Measure and calculate Supply Voltages
  ADMUX = Channel | 0x40;
  // We do one voltage in one time slice
  switch(Channel++) // we get the value from the previous channel
  {
    case 3: AdcMin5A = ADC; // from ch 7 
            Vmin5A = V5A - (V5A - (float(AdcMin5A) * V5D / 1024.0)) * (10.0 + 3.3) / 3.3;
            break;
    case 4: AdcRef = ADC; // from ch 3
            if(AdcRef > 400) V5D = 2500 * 1024.0 / float(AdcRef); else V5D = 0; // prevent division by zero
            break;
    case 5: Adc50 = ADC; // from ch 4 
            V50 = float(Adc50) / 1024.0 * V5D * (44.0 + 3.3) / 3.3;
            break;
    case 6: Adc5A = ADC; // from ch 5
            V5A = float(Adc5A) / 1024.0 * V5D * (1.8 + 3.3) / 3.3;
            break;
    case 7: Adc12 = ADC; // from ch 6 
            V12 = float(Adc12) / 1024.0 * V5D * (9.1 + 3.3) / 3.3;  
            Channel = 3;  
            break;
    default: Channel = 3;
  }
  ADCSRA |= 0x40; // Start new conversion

// Check Supply Voltages within range
  SuppliesOK = true;
  if((V50 < 42000)    || (V50 > 55000))    SuppliesOK = false;  
  if((V12 < 10000)    || (V12 > 13000))    SuppliesOK = false;  
  if((V5A < 4500)     || (V5A > 5500))     SuppliesOK = false; 
  if((Vmin5A < -5500) || (Vmin5A > -4500)) SuppliesOK = false; 
  if(TestTool) SuppliesOK = true;
   
// Check OTW condition
  // CounterOTW is set to 2000 by a Clip_OTW pin change event
  if(CounterOTW > 0) {CounterOTW--; } 
  // If two seconds after a Clip event we still have Clip_OTW low 
  //    we conclude that it is an Over Temperature Warning 
  if(CounterOTW == 1) // 1.999 sec after clip event
  {
    if(ClipOTW_Bpin) OTWB = true;
    if(ClipOTW_MTpin) OTWMT = true;
  }
  // Overtemperature warning is cleared when temperature lowers
  if(!ClipOTW_Bpin) OTWB = false;
  if(!ClipOTW_MTpin) OTWMT = false;  
  
// Evaluate reset of TPA chips
  if(!TestTool) 
  { 
    if(Local) 
    {
      if(!MainsOK || !SuppliesOK || FaultB || FaultMT) RSTTPAON
      else RSTTPAOFF 
    }
   // In remote or on the TestTool it is what the GUI says            
  }

// Check Communication TimeOut
  if(CounterCommunicationTimeout > 0) // Set to 2000 at each incoming message
  { 
    Local = false; 
    CounterCommunicationTimeout--;
  } 
  else 
    Local = true; 
  if(Local) COMledOFF
  
#ifdef FakeLocal
  Local = true;
#endif

// Check condition to talk spontaneously
  if(++CounterSpontaneousMessage > 500) // only 2 x per second
  {
    CounterSpontaneousMessage = 0; 
    if(Local)
      if(!MainsOK || !SuppliesOK || FaultB || FaultMT || OTWB || OTWMT) 
      {
        SendSpontaneousMessage = true;
        SendExtraMessage = true;
      }
      else // Send one "good" message after error is gone
      {
        if(SendExtraMessage) SendSpontaneousMessage = true;
        SendExtraMessage = false;
      }
  }

// Let red ERRORled blink 
  if(ErrorLedReload == -1) ERRORledOFF
  if(ErrorLedReload == -2) ERRORledON
  if(CounterErrorLed > 0) CounterErrorLed--;
  else if(ErrorLedReload > 0)  
  {
    ERRORledTOGGLE;  
    CounterErrorLed = ErrorLedReload;
  }

// Let blue CPUled blink 
  if(MainsOK) CpuLedReload = 1000; else CpuLedReload = 100;
  if(CounterCpuLed > 0)
    CounterCpuLed--;
  else   
  {
    CPUledTOGGLE;  
    CounterCpuLed = CpuLedReload;
  }
  
#ifdef DIAGNOSE_TIMER
  COMledOFF
#endif 
} 

//******************************************************************************************************
void DecodeInMessage(void)
{
  Command = InMessageBuffer[0];
  if(TestTool) {if((Command & 0x10) == 0x10) CrydomON else CrydomOFF } 
  if((Command & 0x20) == 0x20) ClipCountB = 0;  // Clear ClipsB counter
  if((Command & 0x40) == 0x40) ClipCountMT = 0; // Clear ClipsMT counter
  if((Command & 0x80) == 0x80) RSTTPAON else RSTTPAOFF // reset power amplifiers 
  Bas = InMessageBuffer[1];
  Mid = InMessageBuffer[2];
  if(TestTool) return; // skip PGA and EEPROM
  SendDataToPGA(Bas, Mid);
  if((Command & 0x04) == 0x04) // write values to EEPROM
  {
    EEPROM_write(MidAddress, Mid);
    EEPROM_write(BasAddress, Bas);  
  }
}

//******************************************************************************************************
void PrepareOutMessage(void)
{
  int Temp;
  byte Status = 0;
  if(!MainsOK) Status |= 0x01;    // bit 0
  if(!SuppliesOK) Status |= 0x02; // bit 1
  if(FaultB) Status |= 0x04;      // bit 2
  if(FaultMT) Status |= 0x08;     // bit 3
  if(OTWB) Status |= 0x10;        // bit 4   
  if(OTWMT) Status |= 0x20;       // bit 5
  if(!TestTool && SendSpontaneousMessage) Status |= 0x40;  // bit 6
  if(TPAisInReset) Status |= 0x80; // bit 7
  OutMessageBuffer[0] = Status;
  memcpy(&OutMessageBuffer[1], &AdcRef, 2); // ADC value
  Temp = round(V50);                 // in mV
  memcpy(&OutMessageBuffer[3], &Temp, 2); 
  Temp = round (V12);                // in mV
  memcpy(&OutMessageBuffer[5], &Temp, 2);
  Temp = round(V5A);                 // in mV
  memcpy(&OutMessageBuffer[7], &Temp, 2);
  Temp = round(Vmin5A);              // in mV
  memcpy(&OutMessageBuffer[9], &Temp, 2);
  OutMessageBuffer[11] = ClipCountB;
  OutMessageBuffer[12] = ClipCountMT;
}

//********************************************************************************************************
ISR(PCINT0_vect)
// Interrupt from ClipOTW_B on PORTB1 and/or ClipOTW_MT on PORTB3 
{
  static byte PrevPortB;
  CounterOTW = 2000; // 2 seconds ErrorLed
  byte BB = PINB;
  byte B = BB ^ PrevPortB; // exor gives the bits that differ
  PrevPortB = BB;
  if((B & 0x02) == 0x02) {ClipCountB++; SendSpontaneousMessage = true;} 
  if((B & 0x08) == 0x08) {ClipCountMT++; SendSpontaneousMessage = true;}
}

//************************************************************************************************
ISR(PCINT2_vect)
// Interrupt from mains zero crossing on pin D2, PORTD2
{ 
  // pin change interrupt from 50 Hz comes each 10 ms
  CounterMainsOK = 15; // 15 ms timeout  
}

//************************************************************************************************
// The PGA2311 controls volume for Bas en Mid channels in steps of 0.5 dB
// Code 192 is 0dB gain
void SendDataToPGA (byte Bass, byte Midd) // takes 16 usec
// outputs for PGA2311
#define CKlow  PORTC &= ~0x01;
#define CKhigh PORTC |= 0x01;
#define DAlow  PORTC &= ~0x02;
#define DAhigh PORTC |= 0x02;
#define CSlow  PORTC &= ~0x04;
#define CShigh PORTC |= 0x04;
{
  byte i;
  CSlow 
  for(i=0; i<8; i++)
  {
    if ((Midd & 0x80) == 0x80) DAhigh else DAlow
    CKhigh
    Midd = Midd << 1;  
    CKlow
  }
   for(i=0; i<8; i++)
  {
    if ((Bass & 0x80) == 0x80) DAhigh else DAlow
    CKhigh
    Bass = Bass << 1;  
    CKlow
  }
  CShigh 
}

/**************************************************************************************************/
unsigned char EEPROM_read(unsigned int uiAddress)
{
// wait for completion of previous action
while(EECR & (1 << EEPE));
EEAR = uiAddress;
EECR |= (1 << EERE);
return EEDR;
}

/**************************************************************************************************/
void EEPROM_write(unsigned int uiAddress, unsigned char ucData)
{
// wait for completion of previous action
while(EECR & (1<<EEPE));
EEAR = uiAddress;
EEDR = ucData;
EECR |= (1<<EEMPE);
EECR |= (1<<EEPE);
}
 
//******************************************************************************************************
void setup()
{
  Serial.begin(19200);
  Serial.println(VERSIONSTRING);

  Local = true; // we always start in Local mode
  
// configure 16 bit Timer1 for interrupt ISR(TIMER1_COMPA_vect)
  TCCR1A = 0;    // Although PON value = 0 it is required.
  TCCR1B = 0x0A; // select internal clock, prescaler 8, WaveformMode 4
  OCR1A = 1999;  // 1 msec interval = 1000 Hz
  TIMSK1 = 0x02; // interrupt on OCRA match  

// outputs
  DDRD |= 0xE0;  // Leds, will be on after reset
  DDRC |= 0x07;  // PGA lines
  DDRB |= 0x20;  // Led on Nano bord 
  DDRB |= 0x10;  // *Reset for TPA boards

// PullUp's
  PORTB |= 0x0F; // FaultB and ClipOTW pins
  PORTD |= 0x04; // Pull-up for port D2, MainsOK input 
  PORTD |= 0x10; // Pull-up for port D4, test FLT or TestTool 
  
// enable pin change interrupt for PORTB 1 and 3 for ClipOTW_MT inputs
  PCICR |= 0x01;  // PCIE0 for PCINT1 and PCINT3
  PCMSK0 |= 0x0A; // unmask PCINT1 and PCINT3

// enable pin change interrupt for MainsOK input on pin D2
  PCICR |= 0x04;  // PCIE2 for PCINT18 
  PCMSK2 |= 0x04; // unmask PCINT18

// Initialize A/D converter
  ADCSRA = 0x84; // Enable A/D converter, clockdiv 128 for 125 kHz A/D clock
  ADMUX = 0x40;  // Ref = AVCC = +5D
  DIDR0 = 0xF8;  // Disable digital function of analog pins

// load data from EEPROM
  Bas = EEPROM_read (BasAddress);  
  Mid = EEPROM_read (MidAddress);   
  SendDataToPGA(Bas, Mid);  

  Serial.print("Bas= "); Serial.print(Bas);
  Serial.print(" Mid= "); Serial.println(Mid);

// Are we running on FLT or on the TestTool?
// on FLT PORTD4 is GND, on the TestTool it is open
  #ifdef FakeFLT
    TestTool = false;
  #else
    TestTool = ((PIND & 0x10) == 0x10);
  #endif
  Serial.print("Running on ");
  if(TestTool) Serial.println("TestTool_TPA"); else Serial.println("FLT");
  if(TestTool) DDRD |= 0x08; // CR+ on pin D3 is now output
  #ifdef FakeFLT
    Serial.println("Fake FLT ");
  #endif
  #ifdef FakeLocal
    Serial.println("Fake Local ");
  #endif
}

//******************************************************************************************************
void loop()
{
  if(Serial.available() == INMESSAGESIZE)
  {
    #ifndef FakeLocal    
      Local = false; // go into / stay in remote mode
    #endif
    CounterCommunicationTimeout = 2000; // 2 seconds to return to Local when communication stops;
    COMledTOGGLE
    unsigned long wait = millis() + 3;
    while (millis() < wait); // wait until last char received
    Serial.readBytes(InMessageBuffer, INMESSAGESIZE);
    DecodeInMessage();
    PrepareOutMessage();
    Serial.write(OutMessageBuffer, OUTMESSAGESIZE);
    byte dummy[1];  // flush readbuffer to prevent broken messages
    while(Serial.available() > 0) Serial.readBytes(dummy, 1);
  }
  if(SendSpontaneousMessage && !TestTool)
  {
    CounterCommunicationTimeout = 300;
    COMledON
    PrepareOutMessage();
    Serial.write(OutMessageBuffer, OUTMESSAGESIZE);
    SendSpontaneousMessage = false; // only once   
  }
}
